module.exports = function (alchemy) {
    'use strict';

    /**
     * A component system to render static and dynamic CSS
     *
     * @class
     * @name alchemy.ecs.CssRenderSystem
     * @extends alchemy.core.MateriaPrima
     */
    alchemy.formula.define('alchemy.ecs.CssRenderSystem', [
        'alchemy.core.MateriaPrima',
        'alchemy.web.Stylus',

    ], function (MateriaPrima, Stylus) {

        return alchemy.extend(MateriaPrima, {
            /** @lends alchemy.ecs.CssRenderSystem.prototype */

            /** @override */
            constructor: function (cfg) {
                /**
                 * The entity storage
                 *
                 * @property entities
                 * @type alchemy.ecs.Apothecarius
                 * @private
                 */
                this.entities = undefined;

                /**
                 * The css style helper which does the heavy lifting
                 *
                 * @property stylus
                 * @type alchemy.web.Stylus
                 * @private
                 */
                this.stylus = Stylus.brew();

                /**
                 * The the previous state
                 *
                 * @property lastStates
                 * @type Object
                 * @private
                 */
                this.lastStates = {};

                MateriaPrima.constructor.call(this, cfg);
            },

            /** @override */
            dispose: function () {
                this.stylus.dispose();
                this.stylus = null;
                this.entities = null;
                this.lastStates = null;

                MateriaPrima.dispose.call(this);
            },

            /**
             * Updates the component system with the current application state
             */
            update: function () {
                var dynamicCss = this.entities.getAllComponentsOfType('css');
                alchemy.each(dynamicCss, this.updateDynamicCss, this);
            },

            /** @private */
            updateDynamicCss: function (cfg, entityId) {
                this.processTypeRules(cfg, entityId);
                this.processEntityRules(cfg, entityId);
            },

            /** @private */
            processTypeRules: function (cfg, entityId) {
                if (!cfg.typeRules) {
                    return;
                }

                this.setRules(cfg.typeRules);
                this.entities.setComponent(entityId, 'css', {
                    typeRules: null,
                });
            },

            /** @private */
            processEntityRules: function (cfg, entityId) {
                if (!alchemy.isObject(cfg.entityRules)) {
                    this.entities.removeComponent(entityId, 'css');
                    return;
                }

                var rules = {};

                if (alchemy.isFunction(cfg.entityRules)) {
                    var lastState = this.lastStates[entityId];
                    var currentState = this.entities.getComponent(entityId, 'state');

                    if (currentState === lastState) {
                        return;
                    }

                    rules['#' + entityId] = cfg.entityRules.call(null, currentState);

                    this.lastStates[entityId] = currentState;
                    this.setRules(rules);

                    return;
                }

                rules['#' + entityId] = cfg.entityRules;

                this.setRules(rules);
                this.entities.removeComponent(entityId, 'css');
            },

            /** @private */
            setRules: function (rules) {
                this.stylus.setRules(rules);
            },
        });
    });
};
